#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Manage access to the clients, including authenticating when needed."""

import importlib
import logging
import sys

from osc_lib.i18n import _
from osc_lib import clientmanager
from osc_lib import shell
from oslo_config import cfg
import stevedore


LOG = logging.getLogger(__name__)

PLUGIN_MODULES = []
PLUGIN_OPTIONS = {}

USER_AGENT = 'python-openstackclient'


class ClientManager(clientmanager.ClientManager):
    """Manages access to API clients, including authentication

    Wrap osc_lib's ClientManager to maintain compatibility for the existing
    plugin V2 interface.  Some currently private attributes become public
    in osc-lib so we need to maintain a transition period.
    """

    # A simple incrementing version for the plugin to know what is available
    PLUGIN_INTERFACE_VERSION = "2"

    # Let the commands set this
    _auth_required = False

    def __init__(
        self,
        cli_options=None,
        api_version=None,
        pw_func=None,
    ):
        super(ClientManager, self).__init__(
            cli_options=cli_options,
            api_version=api_version,
            pw_func=pw_func,
        )

        # TODO(dtroyer): For compatibility; mark this for removal when plugin
        #                interface v2 is removed
        self._region_name = self.region_name
        self._interface = self.interface
        self._cacert = self.cacert
        self._insecure = not self.verify
        # store original auth_type
        self._original_auth_type = cli_options.auth_type

    def setup_auth(self):
        """Set up authentication"""

        if self._auth_setup_completed:
            return

        # NOTE(dtroyer): Validate the auth args; this is protected with 'if'
        #                because openstack_config is an optional argument to
        #                CloudConfig.__init__() and we'll die if it was not
        #                passed.
        if (
                self._auth_required and
                self._cli_options._openstack_config is not None
        ):
            self._cli_options._openstack_config._pw_callback = \
                shell.prompt_for_password
            try:
                self._cli_options._auth = \
                    self._cli_options._openstack_config.load_auth_plugin(
                        self._cli_options.config,
                    )
            except TypeError as e:
                self._fallback_load_auth_plugin(e)

        return super(ClientManager, self).setup_auth()

    def _fallback_load_auth_plugin(self, e):
        # NOTES(RuiChen): Hack to avoid auth plugins choking on data they don't
        #                 expect, delete fake token and endpoint, then try to
        #                 load auth plugin again with user specified options.
        #                 We know it looks ugly, but it's necessary.
        if self._cli_options.config['auth']['token'] == 'x':
            # restore original auth_type
            self._cli_options.config['auth_type'] = \
                self._original_auth_type
            del self._cli_options.config['auth']['token']
            del self._cli_options.config['auth']['endpoint']
            self._cli_options._auth = \
                self._cli_options._openstack_config.load_auth_plugin(
                    self._cli_options.config,
                )
        else:
            raise e

    def is_network_endpoint_enabled(self):
        """Check if the network endpoint is enabled"""

        # NOTE(dtroyer): is_service_available() can also return None if
        #                there is no Service Catalog, callers here are
        #                not expecting that so fold None into True to
        #                use Network API by default
        return self.is_service_available('network') is not False

    def is_compute_endpoint_enabled(self):
        """Check if Compute endpoint is enabled"""

        return self.is_service_available('compute') is not False

    def is_volume_endpoint_enabled(self, volume_client):
        """Check if volume endpoint is enabled"""
        # NOTE(jcross): Cinder did some interesting things with their service
        #               name so we need to figure out which version to look
        #               for when calling is_service_available()
        volume_version = volume_client.api_version.ver_major
        if self.is_service_available(
                "volumev%s" % volume_version) is not False:
            return True
        elif self.is_service_available('volume') is not False:
            return True
        else:
            return False


# Plugin Support

def _get_config_file_sections(config):
    """Get sections from a configuration file.

    Using oslo.config's cfg.ConfigParser, parses a single configuration file
    into a dictionary containing sections and their respective options.

    The resulting dictionary has the following structure:

    {
        "<section>": {
            "<key>": [<value>, ...],
            ...
        },
        ...
    }

    :param config: the config file
    :returns: dict -- the dictionary of config file sections
    """

    sections = {}

    try:
        cfg.ConfigParser(config, sections).parse()
    except cfg.ParseError as e:
        msg = _(
            'Error while parsing the configuration file `{}`.\n'
            'Location: {}'
        ).format(config, e)

        LOG.error(msg)
    except FileNotFoundError:
        msg = _(
            'No custom config file found for OpenStackClient (OSC). '
            'Using default configurations.'
        ).format(config)
        LOG.debug(msg)

    return sections


def _standardize_list_type_options(options, list_opts):
    """Standardize list-type options.

    With oslo.config's cfg.ConfigParser, depending on how the options are
    specified, their values can be parsed differently. For example:

        load_first = A,B

     Becomes:

        {"load_first": ["A,B"]}

     While:

        load_first = A
        load_first = B

     Becomes:

        {"load_first": ["A", "B"]}

    Therefore, this function standardizes the values of list-type options
    so that they are always presented in the latter format.

    :param options: the options dictionary
    :param list_opts: the list of options to be standardized
    """

    for list_opt in list_opts:
        if list_opt in options:
            values = [
                value.strip()
                for value in ','.join(options[list_opt]).split(',')
                if value.strip()
            ]
            options[list_opt] = values


def process_plugin_options(options):
    """Process plugin-related options.

    :param options: the dictionary of options
    """

    global PLUGIN_OPTIONS

    # If no configuration file was specified,
    # there are no plugin options to process.
    if options.osc_config_file in ['', None]:
        return

    PLUGIN_OPTIONS = _get_config_file_sections(options.osc_config_file).get(
        'plugins', {}
    )

    # Standardizes list-type options' values.
    LIST_OPTS = [
        'load_first',
        'load_last',
    ]

    _standardize_list_type_options(PLUGIN_OPTIONS, LIST_OPTS)


def sort_plugin_entry_points(group):
    """Sort plugin entry points.

    Given a configuration file specified through command-line argument
    (--os-osc-config-file) or environment variable (OS_OSC_CONFIG_FILE),
    this function sorts the plugin entry points based on the `load_first`
    and `load_last` options (if present).

    By setting:

        [plugins]
        load_first = A

    This function will ensure that the plugin "A" is loaded before any other
    plugin. Similarly, by setting:

        [plugins]
        load_last = A

    It will ensure that the plugin "A" is loaded after any other plugin.

    Multiples plugins are supported (as long as they're separated by commas).

    :param group: the group name for the entry points
    :returns: list -- list of entry points

    """

    LOAD_FIRST_PLUGINS = PLUGIN_OPTIONS.get('load_first', [])
    LOAD_LAST_PLUGINS = PLUGIN_OPTIONS.get('load_last', [])

    before_entry_points = []
    after_entry_points = []
    entry_points = []

    mgr = stevedore.ExtensionManager(group)
    for ep in mgr:
        # Different versions of stevedore use different
        # implementations of EntryPoint from other libraries, which
        # are not API-compatible.
        try:
            module_name = ep.entry_point.module_name
        except AttributeError:
            try:
                module_name = ep.entry_point.module
            except AttributeError:
                module_name = ep.entry_point.value

        # If the module name matches at least one plugin specified in
        # `load_first` or `load_last`, append the current entry point
        # to the correspondent list.
        if any([module_name.startswith(cpm) for cpm in LOAD_FIRST_PLUGINS]):
            before_entry_points.append(ep)
        elif any([module_name.startswith(cpm) for cpm in LOAD_LAST_PLUGINS]):
            after_entry_points.append(ep)
        else:
            entry_points.append(ep)

    return before_entry_points + entry_points + after_entry_points


def get_plugin_modules(group):
    """Find plugin entry points"""
    mod_list = []

    for ep in sort_plugin_entry_points(group):
        LOG.debug('Found plugin %s', ep.name)

        # Different versions of stevedore use different
        # implementations of EntryPoint from other libraries, which
        # are not API-compatible.
        try:
            module_name = ep.entry_point.module_name
        except AttributeError:
            try:
                module_name = ep.entry_point.module
            except AttributeError:
                module_name = ep.entry_point.value

        try:
            module = importlib.import_module(module_name)
        except Exception as err:
            sys.stderr.write(
                "WARNING: Failed to import plugin %s: %s.\n" % (ep.name, err))
            continue

        mod_list.append(module)
        init_func = getattr(module, 'Initialize', None)
        if init_func:
            init_func('x')

        # Add the plugin to the ClientManager
        setattr(
            clientmanager.ClientManager,
            module.API_NAME,
            clientmanager.ClientCache(
                getattr(sys.modules[module_name], 'make_client', None)
            ),
        )
    return mod_list


def build_plugin_option_parser(parser):
    """Add plugin options to the parser"""

    # Loop through extensions to get parser additions
    for mod in PLUGIN_MODULES:
        parser = mod.build_option_parser(parser)
    return parser


# Get list of base plugin modules
PLUGIN_MODULES = get_plugin_modules(
    'openstack.cli.base',
)
# Append list of external plugin modules
PLUGIN_MODULES.extend(get_plugin_modules(
    'openstack.cli.extension',
))
